/*
 *	Copyright  2008 George Weller
 *	
 *	This file is part of PathZilla.
 *	
 *	PathZilla is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 2 of the License, or
 *	(at your option) any later version.
 *	
 *	PathZilla is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *	
 *	You should have received a copy of the GNU General Public License
 *	along with PathZilla.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * main.nut
 *
 * PathZilla - A networking AI. See readme.txt for details.
 * 
 * Author:  George Weller (Zutty)
 * Created: 16/05/2008
 * Updated: 08/07/2009
 * Version: 6
 */

class PathZilla extends AIController {
	// Constants - DO NOT change these!
	DIR_NORTH = 1;
	DIR_SOUTH = 2;
	DIR_EAST = 3;
	DIR_WEST = 4;
	TILE_LENGTH_KM = 429;

	// Info constants	
	PZ_IDENT = "ROADRUNNER"
	PZ_VERSION = 3;
	
	// Serialisation constants
	SRLZ_IDENT = 0;
	SRLZ_STOP = 1;
	SRLZ_COMPANY_NAME = 2;
	SRLZ_HOME_TOWN = 3;
	SRLZ_TICKER = 4;
	SRLZ_SCHEMA_IDX = 5;
	SRLZ_SCHEMAS = 6;
	SRLZ_SRVC_MANAGER = 7;
	SRLZ_TRAFFIC_BLACKSPOTS = 8;
	SRLZ_VEHICLES_TO_SELL = 9;
			
	// Configurable constants
	PROCESSING_PRIORITY = 1000;     // Governs how often intensive procesisng tasks should wait
	MAX_TARGETS = 300;             // Maximum number of targets that can be in a single graph 
	MAX_TARGET_COVERAGE = 70;      // Maximum percentage of town houses to fall within combined station coverage area
	NEW_VEHICLE_SPREAD_DELAY = 10; // The delay in ms between launching new vehicles in a fleet.
	MAX_BRIDGE_LENGTH = 15;        // The maximum allowable bridge length - to prevent ridiculous bridges
	MAX_POTENTIAL_SERVICES = 600;  // The maximum allowable number of potential service descriptors  
	ARV_ACC_THRESHOLD = 50;        // Minimum percentage of acceptance via DTRSs before ARVs can be built
	ENGINE_SCORE_THRESHOLD = 80;   // The minimum score for an engine to be randomly selected
	MAX_CONSTR_ATTEMPTS = 8;	   // The maximum number of attempts when trying to build something
	BRIBE_THRESHOLD = 300000;	   // Minimum funds available before a bribe will be considered
	MAX_TREE_SPEND = 10000;		   // Maximum we can spend on trees to improve rating
	MAX_TOWN_RADIUS = 5;		   // Maximum distance from a town centre that anything can be built
	MAX_REPATH_TRIES = 4;		   // Maximum number a times path can be re-found due to construction problems
	MAX_VEHICLES_PER_SVC = 200;	   // Maximum number of vehicles per service
	INDUSTRY_FLEET_MULTI = 4;	   // Fleet size multiplier for industrial services
	TARGET_FIX_RADIUS = 4;		   // Radius around a target we should look to fix a tile
	MAX_INITIAL_STATIONS = 1;	   // Maximum number of stations to start a service with
	PAX_SERVICE_CAP_BASE = 10000000;   // Base level for limit on number of passengers AI will aim to transport
	SERVICE_PROFIT_THRESHOLD = 0;  // Threshold at which to declare a service profitable
	
	// Member variables
	stop = false;
	loaded = false;
	companyName = null;
	homeTown = null;
	schemaIndex = 0;
	schemas = null;
	serviceManager = null;
	ticker = 0;
	FLOAT=5000;
	
	constructor() {
		require("graph/Edge.nut");
		require("graph/Graph.nut");
		require("graph/GraphPathNode.nut");
		require("graph/Triangle.nut");
		require("graph/Vertex.nut");
		require("graph/impl/MinimumSpanTree.nut");
		require("graph/impl/ShortestPathTree.nut");
		require("graph/impl/Triangulation.nut");
		require("manager/FinanceManager.nut");
		require("manager/LandManager.nut");
		require("manager/RoadManager.nut");
		require("manager/TownManager.nut");
		require("pathfinding/PathWrapper.nut");
		require("pathfinding/Road.nut");
		require("schema/Schema.nut");
		require("service/Service.nut");
		require("service/ServiceManager.nut");
		require("service/Target.nut");
		require("struct/Collection.nut");
		require("struct/BinaryHeap.nut");
		require("struct/Map.nut");
		require("struct/SortedSet.nut");
		require("Settings.nut");
		require("common.nut");

		// Some presets that must go here
		this.loaded = false;
		this.schemas = {};
		FLOAT=5000;
		// Set this as the singleton instance
		::pz <- this;
	}
}

/*
 * Start running. Most of the planning, including calculating the plan graph is
 * done before we start looping, though services are selected on the fly. The 
 * main loop manages the loan and events, maintains existing services, attempts
 * to find new services, and then finally builds one.   
 */
function PathZilla::Start() {
::trafficBlackSpots <- AIList();	
::vehiclesToSell <- AIList();	
::german<-false;
AICompany.SetAutoRenewMoney(2000000);
AICompany.SetAutoRenewMonths(4);
AICompany.SetAutoRenewStatus(true);
if (PathZilla.GetSetting("language")==0){
foreach(e,_ in AIEngineList(AIVehicle.VT_ROAD)){
if (AIEngine.GetRoadType(e)==AIRoad.ROADTYPE_ROAD&&AICargo.GetTownEffect(AIEngine.GetCargoType(e))==AICargo.TE_MAIL){
local n=AIEngine.GetName(e);
local str="";
	for (local a=0;a<n.len();a++){
	if ((n[a]).tochar()=="P") str=""
	str=str+(n[a]).tochar();
	if (str=="Postwagen")::german=true;
	}
break;
}
}
}else if(PathZilla.GetSetting("language")==2){
::german=true;
}
(::german)?AILog.Info("  Miep! Miep!"):AILog.Info("  RoadRunner is on the road...");
::major<-((AIController.GetVersion()).tointeger()-(AIController.GetVersion()).tointeger()%268435456).tointeger()/268435456;
::minor<-(((AIController.GetVersion()).tointeger()-(AIController.GetVersion()).tointeger()%16777216).tointeger()-::major*268435456)/16777216;
	// Initialise the AI
this.Initialise();
	//Build headquarter
local searchRadius = PathZilla.MAX_TOWN_RADIUS+10;
		local tileList = AITileList();
		local targetTile=AITown.GetLocation(this.homeTown);
		tileList.AddRectangle(AIMap.GetTileIndex(max(AIMap.GetTileX(targetTile) - searchRadius,1),max(AIMap.GetTileY(targetTile) - searchRadius,1)),AIMap.GetTileIndex(min(AIMap.GetTileX(targetTile) + searchRadius,AIMap.GetMapSizeX()-2),min(AIMap.GetTileY(targetTile) + searchRadius,AIMap.GetMapSizeY()-2)));
		foreach(tile, _ in tileList) {
			local suitable = (AIMap.IsValidTile(tile)&&AITile.IsBuildable(tile) && AITile.GetSlope(tile)==AITile.SLOPE_FLAT&&AITile.IsBuildable(tile+AIMap.GetTileIndex(1,0)) && AITile.GetSlope(tile+AIMap.GetTileIndex(1,0))==AITile.SLOPE_FLAT&&AITile.IsBuildable(tile+AIMap.GetTileIndex(0,1)) && AITile.GetSlope(tile+AIMap.GetTileIndex(0,1))==AITile.SLOPE_FLAT&&AITile.IsBuildable(tile+AIMap.GetTileIndex(1,1)) && AITile.GetSlope(tile+AIMap.GetTileIndex(1,1))==AITile.SLOPE_FLAT);
			tileList.SetValue(tile, (suitable) ? 1 : 0);
		}
		tileList.RemoveValue(0);
		foreach(tile, _ in tileList) {
			local r = AITile.GetDistanceManhattanToTile(tile, AITown.GetLocation(this.homeTown)) + AIBase.RandRange(8) - 4;
			tileList.SetValue(tile, r);
		}
		tileList.Sort(AIAbstractList.SORT_BY_VALUE, true);
if (AICompany.GetCompanyHQ(AICompany.COMPANY_SELF)==AIMap.TILE_INVALID&&!tileList.IsEmpty()){
local tile=tileList.Begin();
while (!AICompany.BuildCompanyHQ(tile)&&hasnext(tileList))tile=tileList.Next();
}
(::german)?AILog.Info(  "Mein Firmensitz ist in "+AITown.GetName(this.homeTown)):AILog.Info("  My home town is " + AITown.GetName(this.homeTown));
	// Initialise the main loop
	local noServices = true;
	local startDate=AIDate.GetCurrentDate();	
	local serlog=AIDate.GetCurrentDate();
	// Load settings for loop latency
	local latency = Settings.GetLatency();
	local workInterval = max(100, latency * 200);
	local maintenanceInterval = max(100, workInterval * latency * 2);
	local expansionInterval = max(20, workInterval * latency * 3);
	// Start the main loop

	while(!this.stop) {
		// Try to keep the amount of funds available around FLOAT, by borrowing
		// or repaying the loan.
		this.FLOAT = (AICompany.GetMaxLoanAmount()/30+AICompany.GetCompanyValue(AICompany.COMPANY_SELF)*AIVehicleList().Count()/10000).tointeger();                 // Minimum amount of money to keep at all times
		FinanceManager.MaintainFunds(PathZilla.FLOAT);
		
		// Check for events
		this.HandleEvents();
if (AIDate.GetCurrentDate()-serlog>366){
serlog=AIDate.GetDate(AIDate.GetYear(serlog)+1,12,31);
local slist=serviceManager.GetServices();
local list=AIList();
local i=0;
foreach (service in slist){
i++;
local profit=0;
if (service.GetActualFleetSize()>0){
foreach (v,_ in service.GetVehicles()){
if(AIVehicle.GetAge(v)>serlog-AIDate.GetDate(AIDate.GetYear(serlog)-1,1,1)) profit+=AIVehicle.GetProfitLastYear(v);
}
if (profit !=0)list.AddItem(i,profit);
}
}
list.Sort(AIAbstractList.SORT_BY_VALUE, false);
list.KeepTop(50);
if (!list.IsEmpty())	((::german)?AILog.Error("   Gruppe   Profit  Fahrzeuge Entfernung"):AILog.Error("   Service   Profit  Vehicles  Distance"));
local iz=0;
foreach(ii,profit in list){
for (profit=profit.tostring()+",-";profit.len()<10;profit+=" ");
iz++;
i=0;
foreach (service in slist){
local st=service+" ";
local vst=(service.GetActualFleetSize()).tostring();
for (;st.len()<31;st+=" ");
for (;vst.len()<10;vst+=" ");
i++;
if (i==ii){	      AILog.Warning(((iz>9)?"":" ")+iz+"."+st+profit+vst+service.GetDistance());
break;
}
}
}
}
		foreach (v,_ in AIVehicleList()){
					if (AIVehicle.IsValidVehicle(v)){
					if (AIVehicle.IsStoppedInDepot(v)||AIVehicle.GetState(v)==AIVehicle.VS_STOPPED){					
					local sl=false;
					foreach(svc in serviceManager.GetServices()){					
					foreach(vs in svc.GetVehicles()){
					if (v==vs){
					sl=true;
					break;
					}
					}
					if (sl)break;
					}
					if (sl){
					AIVehicle.StartStopVehicle(v);
					::vehiclesToSell.RemoveItem(v);
					}
					}
					}
					}
		// Maintain existing services
		if(ticker % maintenanceInterval == 0) {
			this.serviceManager.MaintainServices();
		}
		
		// Look for some new services that we can implement
		this.serviceManager.FindNewServices();
		// Wait until we have a fair bit of cash before building a new line
		if((AIDate.GetCurrentDate()-startDate>300||ticker>300)&&(noServices || (ticker % expansionInterval == 0
			 && FinanceManager.GetAvailableFunds() >= max(60000,AICompany.GetMaxLoanAmount() / 3)))) {
			local vlist=AIVehicleList();
			vlist.Valuate(AIVehicle.GetAge);
			if (vlist.Count()<250||ListSum(vlist)/(vlist.Count()+1)<800||AICompany.GetLoanAmount()<10000){
			this.serviceManager.ImplementService();
			noServices = false;
		}		
		}

		// Advance the ticker
		ticker += workInterval;
//		this.Sleep(workInterval);
	}
}

/*
 * Load state and data structures from a table. This method checks a signature
 * and version number before loading, to ensure that the data being loaded is
 * compatible with this AI. The method also relies on the classes that are used
 * as data structures implementing the Unserialize method.
 */
function PathZilla::Load(version, data) {
	local dataValid = false;
	
	// First check that the data is for this AI, and this version
	if(SRLZ_IDENT in data) {
		if(typeof data[SRLZ_IDENT] == typeof PZ_IDENT) {
			dataValid = (data[SRLZ_IDENT] == PZ_IDENT) && (version == PZ_VERSION);
		}
	}
	
	// If the data is not valid, do not try to load
	if(!dataValid) {
(::german)?AILog.Error("Spielstand benoetigt Version "+ version+" zum Laden."):AILog.Error("Save data was created with version "+version);
		return false;
	}

	this.loaded = true;
	::loadData <- data;
}

/*
 * Save state and data structures to a table for the game to persist. This 
 * method relies on the classes that are used as data structures implementing
 * the Serialize method.
 */
function PathZilla::Save() {
	local data = {};
	
	// Store the ident
	data[SRLZ_IDENT] <- PZ_IDENT;
	
	// Store the global variables
data[SRLZ_TRAFFIC_BLACKSPOTS] <- ListToArray(::trafficBlackSpots);
data[SRLZ_VEHICLES_TO_SELL] <- ListToArray(::vehiclesToSell); 

	// Store the basic data
	data[SRLZ_STOP] <- this.stop;
	data[SRLZ_COMPANY_NAME] <- this.companyName;
	data[SRLZ_HOME_TOWN] <- this.homeTown;
	data[SRLZ_TICKER] <- this.ticker;
	
	// Store the schemas
	data[SRLZ_SCHEMA_IDX] <- this.schemaIndex;
	data[SRLZ_SCHEMAS] <- {};
	foreach(idx, schema in this.schemas) {
		data[SRLZ_SCHEMAS][idx] <- schema.Serialize();
	}

	// Store the service manager, if it has been set
	if(this.serviceManager != null) {
		data[SRLZ_SRVC_MANAGER] <- this.serviceManager.Serialize();
	}
	return data;
}

/*
 * Initialise the state of the AI, either from saved state or from scratch.
 */
function PathZilla::Initialise() {
	// Enable auto-renew
	AICompany.SetAutoRenewStatus(true);
	
	// Set the service manager
	this.serviceManager = ServiceManager();

	// If there is data to load then use it, otherwise start from scratch
	if(this.loaded) {
		// Load some global variables
		::trafficBlackSpots <- ArrayToList(::loadData[SRLZ_TRAFFIC_BLACKSPOTS]); 
		::vehiclesToSell <- ArrayToList(::loadData[SRLZ_VEHICLES_TO_SELL]); 

		// Load the basic data
		this.stop = ::loadData[SRLZ_STOP];
		this.homeTown = ::loadData[SRLZ_HOME_TOWN];
		this.companyName = ::loadData[SRLZ_COMPANY_NAME];

		// Load the schemas
		this.schemaIndex = ::loadData[SRLZ_SCHEMA_IDX];
		foreach(idx, schemaData in ::loadData[SRLZ_SCHEMAS]) {
			this.schemas[idx] <- Schema.instance();
			this.schemas[idx].Unserialize(schemaData);
		}

		// Load the service manager, if it was saved
		if(SRLZ_SRVC_MANAGER in ::loadData) {
			this.serviceManager.Unserialize(::loadData[SRLZ_SRVC_MANAGER]);
		}

		// Load the vehicles into their groups
		this.serviceManager.PostLoad();
	} else {
		// Initialise some global variables
//		::trafficBlackSpots <- AIList();
//		::vehiclesToSell <- AIList();

		// Set the basic data
		this.stop = false;
		this.homeTown = this.SelectHomeTown();
		this.companyName = this.ChooseName();

		// Build the schemas
		this.schemaIndex = -1;
		this.BuildSchemas();
	}
	
	// Set the company name
	AICompany.SetName(trnc(this.companyName));
}

/*
 * Build a series of schemas based on the cargos, towns, and industries 
 * available in the map.
 */
function PathZilla::BuildSchemas() {
	// Add passenger schemas by road and tram
	local townList = AIList();
	local tramList = AIList();
	
	// Check each available cargo
	foreach(cargo, _ in AICargoList()) {
		// Get the amount of this cargo produced in towns
		local townprod = 0;
		foreach(town, _ in AITownList()) {
			townprod += AITown.GetMaxProduction(town, cargo);
		}
		
		// Check if there are any trams that can carry the cargo
		local tramable = false;
		foreach(engine, _ in AIEngineList(AIVehicle.VT_ROAD)) {
			if(AIEngine.GetRoadType(engine) == AIRoad.ROADTYPE_TRAM && AIEngine.CanRefitCargo(engine, cargo)) {
				tramable = true;
				break;
			}
		}
		
		// If the cargo is produced in towns and has a town effect, use it in 
		// the town schema
		if(townprod > 0 && AICargo.GetTownEffect(cargo) != AICargo.TE_NONE) {
			townList.AddItem(cargo, 0);
		}
		
		// If a cargo can be carried by tram and has is of the passenger class,
		// add it to the tram schema
		if(tramable && AICargo.HasCargoClass(cargo, AICargo.CC_PASSENGERS)) {
			tramList.AddItem(cargo, 0);
		}
	}
	
	// Add the town schema
	foreach(c,_ in townList){
	local add=AIList();
	add.AddItem(c,0);	
	this.AddSchema(Schema(this.homeTown, add, AITile.TRANSPORT_ROAD, AIRoad.ROADTYPE_ROAD));
	add=null;
	}
	// Add the tram schema, if they are supported
	if(AIRoad.IsRoadTypeAvailable(AIRoad.ROADTYPE_TRAM)) this.AddSchema(Schema(this.homeTown, tramList, AITile.TRANSPORT_ROAD, AIRoad.ROADTYPE_TRAM));
	
	// Check each available industry type
	foreach(type, _ in AIIndustryTypeList()) {

		// Only add support raw industries taht are not on water
		if(/*AIIndustryType.IsRawIndustry(type) &&*/ !AIIndustryType.IsBuiltOnWater(type)) {
			// Only transport those cargos from this industry that have no town
			// effect, i.e. dont carry passengers from oil rigs, etc...
			local cargos = AIIndustryType.GetProducedCargo(type);
			cargos.Valuate(AICargo.GetTownEffect);
//			cargos.KeepValue(AICargo.TE_NONE);
			
			// Add the schema
	foreach(c,_ in cargos){
local na=true;
local cst=-1;
local d=true;
for(local sc=this.GetNextSchema();;){
if (cst==sc.GetId())break;
if (cst==-1)cst=sc.GetId();
foreach(cg,_ in sc.cargos){
if (cg==c) na=false;
}
}
if (na){
	local add=AIList();
	add.AddItem(c,0);	
	this.AddSchema(Schema(this.homeTown, add, AITile.TRANSPORT_ROAD, AIRoad.ROADTYPE_ROAD));
	add=null;
	}
	}		

	}
	}
}

/*
 * Chooses a company name that does not already exist and returns it. The name
 * must be applied in exec mode separately.
 */
function PathZilla::ChooseName() {
	{
local n=AICompany.GetName(AICompany.COMPANY_SELF);
local str="";
for (local a=0;a<min(n.len(),10);a++){
str=str+(n[a]).tochar();
}
		local _ = AITestMode();
if (str=="RoadRunner") return n;
		local i = 1;
		local name = "";
		
		do {
			if (i==1){
			name = "RoadRunner";
}else{
			name = "RoadRunner #" + i;
}
			i++;
		} while(!AICompany.SetName(trnc(name)));
		
		return name;
	}
}

/*
 * Randomly choose a large town from the top 10 percentile by popuation.
 */
function PathZilla::SelectHomeTown() {
	// Get a list of towns by population
	local towns = AITownList();
	towns.Valuate(AITown.GetPopulation);

	// Remove all but the larges
	local upperLimit = AITown.GetPopulation(towns.Begin());
	local lowerLimit = (upperLimit * 5) / 10;
	towns.RemoveBelowValue(lowerLimit);
	
	// Find towns that have no competitors in them
	foreach(town, _ in towns) {
		// Get a list of tiles to search in
		local townTile = AITown.GetLocation(town);
		local searchRadius = min(AIMap.DistanceFromEdge(townTile) - 1, PathZilla.MAX_TOWN_RADIUS);
		local offset = AIMap.GetTileIndex(searchRadius, searchRadius);
		local tileList = AITileList();
		tileList.AddRectangle(townTile - offset, townTile + offset);
		tileList.Valuate(AITile.IsStationTile);
		tileList.RemoveValue(0);
		towns.SetValue(town, (tileList.IsEmpty()) ? 1 : 0);
	}
	towns.RemoveValue(0);
	
	// If there are no empty towns, just reset the list
	if(towns.IsEmpty()) {
		towns = AITownList();
		towns.Valuate(AITown.GetPopulation);
		towns.RemoveBelowValue(lowerLimit);
	}

	// Select a random town from remaining ones
	towns.Valuate(AIBase.RandItem);
	return towns.Begin();
}

/*
 * Handle any waiting events. This is a place-holder implementation for now!
 */
function PathZilla::HandleEvents(a=false) {
local oz=0;
	while(AIEventController.IsEventWaiting()) {
		local event = AIEventController.GetNextEvent();
		switch(event.GetEventType()) {
			case AIEvent.AI_ET_INDUSTRY_OPEN:
				if (AICompany.GetBankBalance(AICompany.COMPANY_SELF)-AICompany.GetLoanAmount()>-50000&&oz==0&&!a){
				oz=1;
				local open=(AIEventIndustryOpen.Convert(event)).GetIndustryID();
				if (!AIIndustryType.IsBuiltOnWater(AIIndustry.GetIndustryType(open))){
				local log=0
				local ez=0;
				for(;ez<this.schemas.len()-1;ez++) {
				local appschema=GetSchema(ez);
					foreach(c,_ in AIIndustryType.GetProducedCargo(AIIndustry.GetIndustryType(open))) {
						foreach(cargo, _ in appschema.GetCargos()) {
						if (cargo==c) {
						if (log==0)(::german)?AILog.Info(AIIndustry.GetName(open)+" zur Industrieliste hinzugefuegt"):AILog.Info(AIIndustry.GetName(open)+" added to the target list.");
						log=1;
						appschema.InitialiseTargets();
						appschema.Initialise();
						}
						}
					foreach(c,_ in AIIndustryType.GetAcceptedCargo(AIIndustry.GetIndustryType(open))) {
						foreach(cargo, _ in appschema.GetCargos()) {
						if (cargo==c) {
						if (log==0)(::german)?AILog.Info(AIIndustry.GetName(open)+" zur Industrieliste hinzugefuegt"):AILog.Info(AIIndustry.GetName(open)+" added to the target list.");						log=1;
						appschema.InitialiseTargets();
						appschema.Initialise();
						}
						}

					}
					}
					}
					}
				}
			break;
			case AIEvent.AI_ET_INDUSTRY_CLOSE:
				if (!a) break;
				local close=(AIEventIndustryClose.Convert(event)).GetIndustryID();
				local ez=0;
				local services=this.serviceManager.GetServices();
				foreach(service in services) {					
					local tz=0;					
					foreach(target in service.GetTargets()) {					
					if (!target.IsValid()&&service.GetActualFleetSize()>0){
					local v=service.GetVehicles().Begin();
					local ttz=0;
					local station=null;					
					for (local i=0;i<AIOrder.GetOrderCount(v);i++){
					if (AIOrder.IsGotoStationOrder(v,i)){					
					if (ttz==tz) {
					station=AIOrder.GetOrderDestination(v,i);
					break;
					}
					if (station!=AIOrder.GetOrderDestination(v,i)) ttz++;
					}
					}
					local schema=GetSchema(service.GetSchemaId());
					foreach(industry in schema.GetTargets()) {					
					if (station!=null){					
					if (tz==0&&industry.ProducesCargo(service.GetCargo())&&AIMap.DistanceMax(industry.GetLocation(),station)<4){
					service.ChangeTarget(industry.GetId(),tz);
					}else if (tz==1&&industry.AcceptsCargo(service.GetCargo())&&AIMap.DistanceMax(industry.GetLocation(),station)<4){
					service.ChangeTarget(industry.GetId(),tz);
					}					
					}else {
					this.serviceManager.SellVehicles(service,service.GetActualFleetSize());
					this.serviceManager.DeleteService(service);
					}
					}					
					break;					
					}
					tz++;
					}
				}
				foreach(type, _ in AIIndustryTypeList()) {
					if (!AIIndustryType.IsBuiltOnWater(type)) {
					ez++
					local appschema=GetSchema(ez);
					foreach(target in appschema.GetTargets()){
					if (!target.IsValid()){
					this.serviceManager.TargetClosed(target);
//					appschema.InitialiseTargets();
//					appschema.Initialise();
					break;
					}					
					}
					}
				}
				
			case AIEvent.AI_ET_ENGINE_PREVIEW:
				local evt = AIEventEnginePreview.Convert(event);
				evt.AcceptPreview();
			break;
			case AIEvent.AI_ET_COMPANY_IN_TROUBLE:
				if (!a) break;
				local com = AIEventCompanyInTrouble.Convert(event).GetCompanyID();
				if (com==AICompany.COMPANY_SELF){
				serviceManager.SaveMoney();
				}
			break;
			case AIEvent.AI_ET_VEHICLE_WAITING_IN_DEPOT:
				local evt = AIEventVehicleWaitingInDepot.Convert(event);
				local vehicle = evt.GetVehicleID();
				
				// If the vehicle exists and need to be sold, sell it
				if(AIVehicle.IsValidVehicle(vehicle) && ::vehiclesToSell.HasItem(vehicle)) {
					AIVehicle.SellVehicle(vehicle);
					FinanceManager.MaintainFunds(PathZilla.FLOAT);
					::vehiclesToSell.RemoveItem(vehicle);
				}
					
			break;
		}
	}
}

/*
 * Get the netwrok schema with the specified id.
 */
function PathZilla::GetSchema(schemaId) {
	return this.schemas[schemaId];
}

/*
 * Increment the internal schema counter and return the schema with that
 * index. This is used to cycle through schemas in a stateless fashion.
 */
function PathZilla::GetNextSchema() {
	if(++this.schemaIndex >= this.schemas.len()) this.schemaIndex = 0; 
	return this.schemas[this.schemaIndex];
}

/*
 * Add a new network schema to them main table and give it an id.
 */
function PathZilla::AddSchema(schema) {
	local schemaId = this.schemas.len();
	schema.SetId(schemaId);
	return this.schemas[schemaId] <- schema;
	return schemaId;
}
