/*
 *	Copyright  2008 George Weller
 *	
 *	This file is part of PathZilla.
 *	
 *	PathZilla is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 2 of the License, or
 *	(at your option) any later version.
 *	
 *	PathZilla is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *	
 *	You should have received a copy of the GNU General Public License
 *	along with PathZilla.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ServiceManager.nut
 * 
 * Handles all service related functions and maintains service lists.
 * 
 * Author:  George Weller (Zutty)
 * Created: 27/07/2008
 * Version: 1.1
 */

class ServiceManager {
	// Serialization constants
	CLASS_NAME = "ServiceManager";
	SRLZ_SERVICE_LIST = 0;
	SRLZ_TARGETS_UPDATED = 1;
	SRLZ_POTENTIAL_SERVICES = 2;
	SRLZ_TARGETS_CONSIDERED = 3;
	
	// Member variables
	potentialServices = null;
	targetsConsidered = null;
	serviceList = null;
	targetsUpdated = null;

	constructor() {
		this.targetsConsidered = Map();
		this.targetsUpdated = Map();

		this.serviceList = SortedSet();		
		this.potentialServices = SortedSet();
	}
}

/*
 * Get a list of services currently in operation.
 */
function ServiceManager::GetServices() {
	return this.serviceList;
}

function ServiceManager::TargetClosed(target) {
	this.targetsConsidered.Insert(target);
	return;
}

function ServiceManager::DeleteService(svc) {
local helplist=this.serviceList;
this.serviceList=SortedSet();
	foreach(service in helplist) {
if (service.GetCargo()!=svc.GetCargo()||service.GetTargets()[0]!=svc.GetTargets()[0]||service.GetTargets()[1]!=svc.GetTargets()[1]){
this.serviceList.Insert(service);
}
}
	return;
}

/*
 * Maintians the operating services. This function ensures that all towns have
 * enough stations to cover them, and that all the vehciles in each service are
 * distributed adequately.
 */
function ServiceManager::MaintainServices() {
	local targetsTried = SortedSet();
	foreach(service in this.serviceList) {
//		PathZilla.Sleep(1);
		
		// Update fleet size
		this.CreateFleet(service, true);

		if(service.GetTransportType() == AITile.TRANSPORT_ROAD&&AIDate.GetCurrentDate()%20==1) {
			RoadManager.MaintainInfrastructure(service, targetsTried, this.targetsUpdated);
		}
		
		local needUpdate = false;
		foreach(target in service.GetTargets()) {
			needUpdate = needUpdate && this.targetsUpdated.Contains(target);
		}
		if(needUpdate) {
(::german)?AILog.Info("  Ueberpruefe: " + service):AILog.Info("  Updating service - " + service);
			this.UpdateOrders(service);
		}
	}

	this.targetsUpdated = Map();
}

/*
 * Gets a list of potential services as service descriptors. At present this
 * searchs through a matrix of all towns in the map, checks which would turn a
 * profit, and ranks them by their profitability.
 */
function ServiceManager::FindNewServices() {
	local schema = ::pz.GetNextSchema();
//	local schema = ::pz.GetSchema(0);
	local transportType = schema.GetTransportType();
	local subType = schema.GetSubType();
	// If there are no targets then just move on
if (schema.GetTargets().len()<2)return;
	// Discard the targets that we have already been to, or that can't be reached
	this.targetsConsidered.Insert((schema.GetTargets())[schema.GetTargets().len()-1]);
local spd=168
local vlist=AIVehicleList();
if (!vlist.IsEmpty()){
vlist.Valuate(AIVehicle.GetCurrentSpeed);
vlist.Sort(AIAbstractList.SORT_BY_VALUE, false);
spd=max(spd,(vlist.GetValue(vlist.Begin()))*3);
}
local tlist=AIList();
foreach(t in schema.GetTargets()){
if (!t.IsTown()||AITown.GetPopulation(t.GetId()-1000000)>500)tlist.AddItem(t.GetId(),0);
}
local targetList = [];
if (((AICompany.GetBankBalance(AICompany.COMPANY_SELF)/10000).tointeger())%50==20){
this.targetsConsidered.RemoveAll(this.targetsConsidered);
			this.targetsConsidered.Insert(schema.GetTargets()[0]);
}
foreach (tk in schema.GetTargets()){
local insert=0;
if (this.targetsConsidered!=null){
foreach (tkk in this.targetsConsidered.GetData()){
if (tk.GetName()==tkk.GetName()){
insert=1;
break;
}
}
}
if (insert==0&&tk.IsValid()&&(!tk.IsTown()||AITown.GetPopulation(tk.GetId()-1000000)>500)){
 targetList.append(tk);
}
}
	// If all targets have already been considered then just move on
	if(targetList.len() == 0) return;
	// Look for possible services for each cargo type in the current schema
local lcar=0;
foreach(cargo, _ in schema.GetCargos()) {
lcar++;
}
local lzz=0
local cargs=schema.GetCargos();
	for(local cargo=cargs.Begin();hasnext(cargs);cargo=cargs.Next()){
lzz++;
(::german)?0:AILog.Info("  Looking for potential " + AICargo.GetCargoLabel(cargo) + " services...");
if (subType==AIRoad.ROADTYPE_TRAM){
local eg=0;
local en=null;
foreach(e,_ in AIEngineList(AIVehicle.VT_ROAD)){
if (AIEngine.GetCargoType(e)==cargo){
eg=max(eg,AIEngine.GetMaxSpeed(e));
en=e;
}
}
if (en==null) return;
if (AIEngine.GetRoadType(en)==AIRoad.ROADTYPE_ROAD) subType=AIRoad.ROADTYPE_ROAD;
}
		// Discard those targets that don't produce this cargo
local engine = this.SelectEngine(targetList, cargo, transportType, subType, false);
		if(engine != null) {
		local targets = [];
foreach (tk in schema.GetTargets()){
if (tk.IsValid()&&tk.ProducesCargo(cargo)){
local trz=0;
foreach (tr in targets){
if (tr.IsTown()){
if ((AITown.GetMaxProduction(tk.GetId()-1000000,cargo)/((tk.GetTransportation(-1,cargo)+49)/50+1).tointeger())*(100-tk.GetTransportation(-1,cargo)+1)>=(AITown.GetMaxProduction(tr.GetId()-1000000,cargo)/((tr.GetTransportation(-1,cargo)+49)/50+1).tointeger())*(100-tr.GetTransportation(-1,cargo)+1)){
break;
}
}else if ((AIIndustry.GetLastMonthProduction(tk.GetId(),cargo)/((tk.GetTransportation(-1,cargo)+49)/50+1).tointeger())*(100-tk.GetTransportation(-1,cargo)+1)>=(AIIndustry.GetLastMonthProduction(tr.GetId(),cargo)/((tr.GetTransportation(-1,cargo)+49)/50+1).tointeger())*(100-tr.GetTransportation(-1,cargo))){
break;
}
trz++;
}
/*if (tk.IsTown()){AILog.Info(AITown.GetMaxProduction(tk.GetId()-1000000,cargo)+AICargo.GetCargoLabel(cargo)+tk.GetTransportation(-1,cargo)+"%,"+((tk.GetTransportation(-1,cargo)+49)/50+1).tointeger()+","+(100-tk.GetTransportation(-1,cargo))+tk.GetName()+"="+(AITown.GetMaxProduction(tk.GetId()-1000000,cargo)/((tk.GetTransportation(-1,cargo)+49)/50+1).tointeger())*(100-tk.GetTransportation(-1,cargo)+1)+"|"+trz);
}else{AILog.Info(AIIndustry.GetLastMonthProduction(tk.GetId(),cargo)+AICargo.GetCargoLabel(cargo)+tk.GetTransportation(-1,cargo)+"%,"+((tk.GetTransportation(-1,cargo)+49)/50+1).tointeger()+","+(100-tk.GetTransportation(-1,cargo))+tk.GetName()+"="+(AIIndustry.GetLastMonthProduction(tk.GetId(),cargo)/((tk.GetTransportation(-1,cargo)+49)/50+1).tointeger())*(100-tk.GetTransportation(-1,cargo)+1)+"|"+trz);
}
*/
if (trz<targets.len()){
 targets[trz]=tk;
}else{
targets.append(tk);
}
}
}
foreach (tk in schema.GetTargets()){
if (tk.IsValid()&&tk.ProducesCargo(cargo)) targets.append(tk);
}
		
for (local it=0;it<min(10,targets.len()-1);it++){
			local aTarget=targets[(it+(AIDate.GetYear(AIDate.GetCurrentDate())%20)*10)%(targets.len()-1)];
if (lcar==lzz) this.targetsConsidered.Insert(aTarget);
(::german)?AILog.Info("    Suche moegliche Ziele fuer "+AICargo.GetCargoLabel(cargo)+" von " + aTarget.GetName() + "..."):AILog.Info("    Looking for potential services from " + aTarget.GetName() + "...");
				// Get the shortest distances accross the network
			local netDist = schema.GetPlanGraph().GetShortestDistances(aTarget.GetVertex());
			// Iterate over each town to test each possible connection
			local steps = 0;
			local blist= tlist;
 			blist.Valuate(function(b,a,spd){return abs(AIMap.DistanceManhattan(((b>1000000)?AITown.GetLocation(b-1000000):AIIndustry.GetLocation(b)),a)-spd)},aTarget.GetLocation(),spd);
			blist.Sort(AIAbstractList.SORT_BY_VALUE, true);
			blist.KeepTop(15);
			foreach(b,_ in blist) {
local bTarget=((b>1000000)?Target(Target.TYPE_TOWN, b):Target(Target.TYPE_INDUSTRY, b));
				if(bTarget.AcceptsCargo(cargo)){
					// Build a list of targets
				local targetIds = [aTarget.GetId(), bTarget.GetId()];
				local targetList = [aTarget, bTarget];

				// Ensure that its possible to connect to the town, and that we 
				// don't already provide this service
				if(bTarget != aTarget && !this.ProvidesService(targetIds, cargo, transportType, subType)) {
					local bTile = bTarget.GetLocation();
	
					// Select an engine
					local crowDist = AITile.GetDistanceManhattanToTile(aTarget.GetLocation(), bTile);
					if (bTile in netDist){
					if(crowDist>20&&netDist[bTile]*2<crowDist*3&& netDist[bTile] > 0) {
					
					// Decide on a limit for the target coverage level
					local coverageLimit = PathZilla.MAX_TARGET_COVERAGE;
					local year = AIDate.GetYear(AIDate.GetCurrentDate());
					if(year < 1950) {
						year = max(year, 1910);
						coverageLimit = ((year - 1900) * 16 / 10);
					}
	
					// Decide on the coverage level itself
					// TODO: Move this code into the schema and make it more general
					local maxCoverage = PathZilla.MAX_TARGET_COVERAGE;
					local coverageTarget = maxCoverage;
					if(subType == AIRoad.ROADTYPE_TRAM) coverageTarget = maxCoverage / 2; // Penalise trams to prevent sprawl
					if(AICargo.HasCargoClass(cargo, AICargo.CC_MAIL)) coverageTarget = maxCoverage / 4; // Penalise mail to prevent over-servicing
					
					// Ensure the target does not exceed the limit
					coverageTarget = min(coverageTarget, coverageLimit);
					
					// Only consider the service if it is more profitable than it is costly
					local travelTime=((netDist[bTile]*30)/(sqrt(AIEngine.GetMaxSpeed(engine)+1)*4*((netDist[bTile]+15)/(netDist[bTile]+1)))).tointeger(); // in days
					// Get the base income for one trip				
					local rawIncome = AICargo.GetCargoIncome(cargo, crowDist, travelTime)*AIEngine.GetCapacity(engine);
					local production = (aTarget.IsTown()?(AITown.GetMaxProduction(aTarget.GetId()-1000000,cargo)/((aTarget.GetTransportation(-1,cargo)+49)/50+1).tointeger())*(100-aTarget.GetTransportation(-1,cargo)+1)/100:(AIIndustry.GetLastMonthProduction(aTarget.GetId(),cargo)/((aTarget.GetTransportation(-1,cargo)+49)/50+1).tointeger())*(100-aTarget.GetTransportation(-1,cargo)+1)/100).tointeger();
if (Service(schema.GetId(), targetIds, cargo, transportType, subType, engine, crowDist, rawIncome, coverageTarget).IsTwoWayService()){
local productionb = (((bTarget.IsTown()?(AITown.GetMaxProduction(bTarget.GetId()-1000000,cargo)/((bTarget.GetTransportation(-1,cargo)+49)/50+1).tointeger())*(100-bTarget.GetTransportation(-1,cargo)+1)/100:(AIIndustry.GetLastMonthProduction(bTarget.GetId(),cargo)/((bTarget.GetTransportation(-1,cargo)+49)/50+1).tointeger())*(100-bTarget.GetTransportation(-1,cargo)+1)/100))).tointeger();
production=((production+productionb-max(productionb-production,production-productionb)/3*2)*2).tointeger();
}					// Project revenue and costs
					local annualProfit = ((rawIncome-((travelTime+20)*AIEngine.GetRunningCost(engine)/180))*production*365/(travelTime+20)).tointeger();
if ((this.potentialServices.Len()==0||annualProfit*8>(this.potentialServices.Begin()).GetRawIncome())){
local water=1000;
local stpo=max(abs(AIMap.GetTileX(aTarget.GetLocation())-AIMap.GetTileX(bTarget.GetLocation())),abs(AIMap.GetTileY(aTarget.GetLocation())-AIMap.GetTileY(bTarget.GetLocation())));
local wz=0;
for (local i=2;i<=stpo&&wz<=10000000;i++){
local wx=(AIMap.GetTileX(aTarget.GetLocation())-(AIMap.GetTileX(aTarget.GetLocation())-AIMap.GetTileX(bTarget.GetLocation()))*i/stpo).tointeger();
local wy=(AIMap.GetTileY(aTarget.GetLocation())-(AIMap.GetTileY(aTarget.GetLocation())-AIMap.GetTileY(bTarget.GetLocation()))*i/stpo).tointeger();
//if (i%10==0)AILog.Info(wx+"/"+wy+"|"+wz);
if (AITile.IsWaterTile(AIMap.GetTileIndex(wx,wy))){
wz=wz+1+(sqrt(wz*7));
}else{
if (wz>0) water=water+wz;
wz=0;
}
}
local wa=1000;
local wz=0;
for (local i=1;i<=AIMap.DistanceManhattan(aTarget.GetLocation(),bTarget.GetLocation())&&wz<=10000000;i++){
local wx=(AIMap.GetTileX(aTarget.GetLocation())-((AIMap.GetTileX(aTarget.GetLocation())>AIMap.GetTileX(bTarget.GetLocation()))?min((AIMap.GetTileX(aTarget.GetLocation())-AIMap.GetTileX(bTarget.GetLocation())),i):max((AIMap.GetTileX(aTarget.GetLocation())-AIMap.GetTileX(bTarget.GetLocation())),0-i)));
local wy=(AIMap.GetTileY(aTarget.GetLocation())-((AIMap.GetTileY(aTarget.GetLocation())>AIMap.GetTileY(bTarget.GetLocation()))?min((AIMap.GetTileY(aTarget.GetLocation())-AIMap.GetTileY(bTarget.GetLocation())),max(0,i-abs(AIMap.GetTileX(aTarget.GetLocation())-AIMap.GetTileX(bTarget.GetLocation())))):max((AIMap.GetTileY(aTarget.GetLocation())-AIMap.GetTileY(bTarget.GetLocation())),0-max(0,i-abs(AIMap.GetTileX(aTarget.GetLocation())-AIMap.GetTileX(bTarget.GetLocation()))))));
if (AITile.IsWaterTile(AIMap.GetTileIndex(wx,wy))){
wz=wz+1+(sqrt(wz*7));
}else{
if (wz>0) wa=wa+wz;
wz=0;
}
}
water=min(water,wa);
wa=1000;
wz=0;
for (local i=1;i<=AIMap.DistanceManhattan(aTarget.GetLocation(),bTarget.GetLocation())&&wz<=10000000;i++){
local wy=(AIMap.GetTileY(aTarget.GetLocation())-((AIMap.GetTileY(aTarget.GetLocation())>AIMap.GetTileY(bTarget.GetLocation()))?min((AIMap.GetTileY(aTarget.GetLocation())-AIMap.GetTileY(bTarget.GetLocation())),i):max((AIMap.GetTileY(aTarget.GetLocation())-AIMap.GetTileY(bTarget.GetLocation())),0-i)));
local wx=(AIMap.GetTileX(aTarget.GetLocation())-((AIMap.GetTileX(aTarget.GetLocation())>AIMap.GetTileX(bTarget.GetLocation()))?min((AIMap.GetTileX(aTarget.GetLocation())-AIMap.GetTileX(bTarget.GetLocation())),max(0,i-abs(AIMap.GetTileY(aTarget.GetLocation())-AIMap.GetTileY(bTarget.GetLocation())))):max((AIMap.GetTileX(aTarget.GetLocation())-AIMap.GetTileX(bTarget.GetLocation())),0-max(0,i-abs(AIMap.GetTileY(aTarget.GetLocation())-AIMap.GetTileY(bTarget.GetLocation()))))));
if (AITile.IsWaterTile(AIMap.GetTileIndex(wx,wy))){
wz=wz+1+(sqrt(wz*7));
}else{
if (wz>0) wa=wa+wz;
wz=0;
}
}
water=min(water,wa);
if (!aTarget.IsTown())annualProfit=annualProfit/max(1,(AIIndustry.GetAmountOfStationsAround(aTarget.GetId())-RoadManager.GetStations(aTarget, cargo, subType).Count()));
annualProfit=((annualProfit*1000)/water).tointeger();
//AILog.Info(aTarget.GetName()+"-"+bTarget.GetName()+crowDist+"dist."+travelTime+"days"+annualProfit+"$"+water);
						local svc = Service(schema.GetId(), targetIds, cargo, transportType, subType, engine, crowDist, annualProfit, coverageTarget);
this.potentialServices.Insert(svc);
					}					
					}else{	
//	AILog.Error("    There is no possible path between " + aTarget.GetName() + " and " + bTarget.GetName());
					}
					}
					}
					}
					}
					}
					
			// To prevent an exponential buildup of descriptors, keep only the top 
			// MAX_POTENTIAL_SERVICES most profitable ones
if (this.potentialServices.Len()>PathZilla.MAX_POTENTIAL_SERVICES){
local g=0;
this.potentialServices.SortBy(profit);
local svs=this.potentialServices;
for (local pi=0;pi<svs.Len();pi++){
local sv=this.potentialServices[pi];
//AILog.Warning(sv+"|"+sv.rawIncome);
if (g++>PathZilla.MAX_POTENTIAL_SERVICES||sv.GetRawIncome()*10<PathZilla.FLOAT||sv.GetDistance()<20) this.potentialServices.Remove(sv);
}
}
this.potentialServices.SortBy(profit);
(::german)?AILog.Info("    Fertig."):AILog.Info("    Done.");
		}

	}
}

/*
 * Checks to see if the company provides a service from a to b for the
 * specified cargo and road type.
 */
function ServiceManager::ProvidesService(targetIds, cargo, transportType, subType) {
	foreach(service in this.serviceList) {
		if(service.GetCargo() == cargo && service.GoesToAll(targetIds) && service.GetTransportType() == transportType && service.GetSubType() == subType) {
			return true;
		}
	}
	
	return false;
}

/*
 * Checks to see if the company already provides the specified service.
 */
function ServiceManager::ProvidesThisService(svc) {
	return this.ProvidesService(svc.GetTargetIds(), svc.GetCargo(), svc.GetTransportType(), svc.GetSubType());
}

/*
 * Choose the next best service descriptor from pententialServices to be 
 * implemented in the game world. This function ensures that roads are built to
 * connect the towns in the service to the network (described by actualGraph),
 * builds preliminary stations at the town, and then creates the vehciles that
 * will operate the service.
 */
function ServiceManager::ImplementService() {
	// Check whether or not we can build any more vehicles
	if(AIVehicleList().Count() == AIGameSettings.GetValue("vehicle.max_roadveh")) return false;
	
	// Implement the service at the top of the list

	local bestService = null;
for (local pi=0;pi<this.potentialServices.Len();pi++){
bestService=this.potentialServices[pi];
local cont=0;
	// Check that we don't already provide this service
for(local tw=0;(tw<1||(bestService.GetTargets()[0]).IsTown())&&tw<2;tw++){
cont=0;
if (tw>0&&(bestService.GetTargets()[0]).IsTown()){
if (bestService.GetCargo()!=0){
local engine = this.SelectEngine(bestService.GetTargets(), 0, bestService.GetTransportType(), bestService.GetSubType(), false);
if (engine==null) continue;
bestService=Service(bestService.GetSchemaId(),bestService.GetTargetIds(),0,bestService.GetTransportType(),bestService.GetSubType(),engine,bestService.GetDistance(),bestService.GetRawIncome(),bestService.GetCoverageTarget());
}
}
local ll=0;
		// If we already provide it then move on to the next one
if (!this.ProvidesThisService(bestService)&&bestService.WasTransported()>=(bestService.GetTargets()[0]).GetTransportation(-1,bestService.GetCargo())/((bestService.GetTargets()[0].IsTown())?1:max(1,(AIIndustry.GetAmountOfStationsAround(bestService.GetTargetIds()[0])-RoadManager.GetStations(bestService.GetTargets()[0], bestService.GetCargo(), bestService.GetSubType()).Count())))){
foreach(serv in this.serviceList) {
if (serv.GetDistance()>1&&bestService.GetCargo()==serv.GetCargo()&&((serv.GetTargets())[0]).GetName()==((bestService.GetTargets())[0]).GetName()){
ll=1;
break;
}
//AILog.Warning(serv.GetDistance()+"|"+ll+"|"+bestService.GetCargo()+"="+serv.GetCargo()+"|"+((serv.GetTargets())[0]).GetName()+"="+((bestService.GetTargets())[0]).GetName());	
}
}else{ ll=1;
}
if (ll==1) {
cont=1;
break
}
	
	// Only proceed if there are any services left to implement
	if(bestService != null) {
		local success = false;
		local schema = ::pz.GetSchema(bestService.GetSchemaId());
		
(::german)?AILog.Info("Beste Verbindung liefert " + bestService+"."):AILog.Info("Best service takes " + bestService);
		
		// Build infrastructure for the service
		if(bestService.GetTransportType() == AITile.TRANSPORT_ROAD) {
			success = RoadManager.BuildInfrastructure(bestService, schema, this.targetsUpdated);
		}
		
		// If the service implementation failed then move it from the top 
		// position in the list to allow other services to be implemented
		if(!success) {
(::german)?AILog.Warning("  Verwerfe geplante Route..."):AILog.Warning("  Demoting service...");
			this.potentialServices.Remove(bestService);
			this.potentialServices.SortBy(profit);
			cont=1;
			break;
		}

		// Create a fleet of vehicles to operate this service
if (cont==0){
		bestService.Create();
		this.CreateFleet(bestService);
}
		// Finally, add the service to the list	
if (bestService.GetActualFleetSize()<1){
(::german)?AILog.Warning("  Keine Fahrzeuge vorhanden..."):AILog.Warning("  No fleet created...");
			this.potentialServices.Remove(bestService);
			this.potentialServices.SortBy(profit);
			cont=1;
			break;
		}
		this.serviceList.Insert(bestService);
local cargoList=AIList();
cargoList.AddItem(bestService.GetCargo(),0);
if (::german){
AILog.Warning("Service Nr:     "+bestService.GetSchemaId());
AILog.Warning("Fracht:            "+AICargo.GetCargoLabel(bestService.GetCargo()));
AILog.Warning("Von:                    "+((bestService.GetTargets())[0]).GetName());
AILog.Warning("Nach:                "+((bestService.GetTargets())[1]).GetName());
AILog.Warning("Entfernung: "+bestService.GetDistance());
AILog.Warning("Einkommen: "+bestService.GetRawIncome());
}else{
AILog.Warning("Service Nr:     "+bestService.GetSchemaId());
AILog.Warning("Cargo:              "+AICargo.GetCargoLabel(bestService.GetCargo()));
AILog.Warning("From:                  "+((bestService.GetTargets())[0]).GetName());
AILog.Warning("To:                       "+((bestService.GetTargets())[1]).GetName());
AILog.Warning("Distance:        "+bestService.GetDistance());
AILog.Warning("Raw Income:"+bestService.GetRawIncome());
}
if (!(bestService.GetTargets()[0]).IsTown()&&!(bestService.GetTargets()[1]).IsTown()&&bestService.IsValid()){
		foreach (prodcargo , _ in AIIndustryType.GetProducedCargo(AIIndustry.GetIndustryType((bestService.GetTargets()[0]).GetId()))){
		foreach (acceptcargo , _ in AIIndustryType.GetAcceptedCargo(AIIndustry.GetIndustryType((bestService.GetTargets()[1]).GetId()))){
		if (!cargoList.HasItem(prodcargo)&&prodcargo==acceptcargo&&AICargo.GetTownEffect(prodcargo)!=AICargo.TE_PASSENGERS&&AICargo.GetTownEffect(prodcargo)!=AICargo.TE_MAIL) {
local engine = this.SelectEngine(bestService.GetTargets(), prodcargo, bestService.GetTransportType(), bestService.GetSubType(), false);
local farmService=Service(bestService.GetSchemaId(),bestService.GetTargetIds(),prodcargo,bestService.GetTransportType(),bestService.GetSubType(),engine,bestService.GetDistance(),bestService.GetRawIncome(),bestService.GetCoverageTarget());
local schema = ::pz.GetSchema(farmService.GetSchemaId());
		farmService.Create();
		this.CreateFleet(farmService);
                this.serviceList.Insert(farmService);
if (::german){
AILog.Warning("Service Nr:     "+bestService.GetSchemaId());
AILog.Warning("Fracht:            "+AICargo.GetCargoLabel(prodcargo));
AILog.Warning("Von:                    "+((bestService.GetTargets())[0]).GetName());
AILog.Warning("Nach:                "+((bestService.GetTargets())[1]).GetName());
AILog.Warning("Entfernung: "+bestService.GetDistance());
AILog.Warning("Einkommen: "+bestService.GetRawIncome());
}else{
AILog.Warning("Service Nr:     "+bestService.GetSchemaId());
AILog.Warning("Cargo:              "+AICargo.GetCargoLabel(prodcargo));
AILog.Warning("From:                  "+((bestService.GetTargets())[0]).GetName());
AILog.Warning("To:                       "+((bestService.GetTargets())[1]).GetName());
AILog.Warning("Distance:        "+bestService.GetDistance());
AILog.Warning("Raw Income:"+bestService.GetRawIncome());
}
cargoList.AddItem(prodcargo,0);
}
}
}
}
}
if (cont==1) {
			this.targetsConsidered.Remove(best.Service.GetTargets()[0]);
continue;
}
(::german)?AILog.Info("Verbindung erfolgreich fertiggestellt."):AILog.Info("Done implementing service.");
	}

	// Don't remove it until we are finished
			this.potentialServices.Remove(bestService);
			this.potentialServices.SortBy(profit);
if (cont==1&&FinanceManager.GetAvailableFunds() >= (AICompany.GetMaxLoanAmount() / 4)) continue;
		return true;
}
}

/*
 * Choose an engine to run between two specified towns, and carry the specified
 * cargo. This method is compatible with NewGRF sets that require vehciles to 
 * be refitted.
 */
function ServiceManager::SelectEngine(targets, cargo, transportType, subType, checkStations) {
	local availableFunds = FinanceManager.GetAvailableFunds();
	
	local vtTypeMap = {};
	vtTypeMap[AITile.TRANSPORT_ROAD] <- AIVehicle.VT_ROAD;
	
	local engineList = AIEngineList(vtTypeMap[transportType]);
	foreach(engine, _ in engineList) {
		local ok = true; 
		if(transportType == AITile.TRANSPORT_ROAD) {
			if(AIEngine.GetRoadType(engine) != subType) ok = false;
		}
		if(!(AIEngine.GetCargoType(engine) == cargo || AIEngine.CanRefitCargo(engine, cargo))) ok = false;
		if(AIEngine.GetPrice(engine) == 0) ok = false;
		if(AIEngine.GetPrice(engine) > availableFunds) ok = false;
		if(AIEngine.GetCapacity(engine) <= 0) ok = false;
		engineList.SetValue(engine, (ok) ? 1 : 0);
	}
	
	// Discount vehicles that are invalid or that can't be built
	engineList.RemoveValue(0);

	// If none are left, then return with nothing	
	if(engineList.Count() == 0) {
		return null;
	}
	
	// Calculate the total distance
	local distance = 0;
	local prev = 0;
	for(local next = 1; next < targets.len(); next++) {
		distance += AITile.GetDistanceManhattanToTile(targets[prev].GetTile(), targets[next].GetTile());
	}
		
	// Build a function to compute the profit making potential of each vehicle
	local profitValuator = function (engine, cargo, distance) {
		local travelTime = (179 * distance) / (10 * AIEngine.GetMaxSpeed(engine)); // AIEngine.GetReliability(engine) / 100
		travelTime = max(1, travelTime); // Compensate for ultra-fast vehicles
		local unitIncome = AICargo.GetCargoIncome(cargo, distance, travelTime);
		local period = 5; // years
		local tco = AIEngine.GetPrice(engine) + (AIEngine.GetRunningCost(engine) * period);
		local income = unitIncome * AIEngine.GetCapacity(engine) * ((364 * period) / travelTime); 
		local profit = income - tco;
		return profit;

	}
	
	// Find the highest profit level
	engineList.Valuate(profitValuator, cargo, distance);
	local maxProfit = engineList.GetValue(engineList.Begin());
	
	// Findthe highest capacity
	engineList.Valuate(AIEngine.GetCapacity);
	local maxCapactiy = engineList.GetValue(engineList.Begin());
	
	// Get the minimum acceptance for the service
	local minAcceptance = -1;
	
	if(checkStations) { 
		// Get coverage radius of stations the vechies will stop at
		local truckStation = !AICargo.HasCargoClass(cargo, AICargo.CC_PASSENGERS);
		local stationType = (truckStation) ? AIStation.STATION_TRUCK_STOP : AIStation.STATION_BUS_STOP;
		local radius = AIStation.GetCoverageRadius(stationType);

		// Create a valuator function to rank stations based on acceptance
		local accValuator = function(station, cargo, radius) {
			return AITile.GetCargoAcceptance(AIStation.GetLocation(station), cargo, 1, 1, radius) + 1;
		}

		// Get the minimum level of acceptance for each target		
		minAcceptance = 10000000;
		foreach(target in targets) {
			local stations = RoadManager.GetStations(target, cargo, subType);
			stations.Valuate(accValuator, cargo, radius);
			minAcceptance = min(minAcceptance, ListSum(stations));
		}
	}
	
	// Rank the remaining engines by their score
	foreach(engine, _ in engineList) {
		local profitTerm = (max(0, profitValuator(engine, cargo, distance)) * 100) / maxProfit;
		local reliabilityTerm = AIEngine.GetReliability(engine);
		local normCapacity = (AIEngine.GetCapacity(engine) * 100) / maxCapactiy;
		local accUpper = 250;
		local overkillTerm = 100 - abs(normCapacity - (min(minAcceptance, accUpper) * 100 / accUpper));
		local score = (profitTerm + reliabilityTerm + overkillTerm) / 3;
		engineList.SetValue(engine, score);
	}
	engineList.Sort(AIAbstractList.SORT_BY_VALUE, false);
	
	// If the engines are good enough then choose randomly from the best ones 
	if(engineList.GetValue(engineList.Begin()) >= PathZilla.ENGINE_SCORE_THRESHOLD) {
		engineList.RemoveBelowValue(PathZilla.ENGINE_SCORE_THRESHOLD);
		engineList.Valuate(AIBase.RandItem);
	}
	
	// Return the selected engine
	return engineList.Begin();
}

/*
 * Create a fleet of vehicles for the specified service. This method assumes 
 * that the towns that the service run between are already on the network and  
 * have stations built. The function finds the nearest depot, then estimates a 
 * suitable fleet size, then builds the vehicles with randomly distributed 
 * orders between the stations in both towns.
 */
function ServiceManager::CreateFleet(service, update = false) {
	// Initialise
	local vnum=((AIVehicleList().Count())/100+1).tointeger();
        local cargo = service.GetCargo();	
        local isIndustry = false;
	// Get the stations
	local stations = {};
	local depots = {};
	// Get type of station the vechies will stop at
	local truckStation = !AICargo.HasCargoClass(cargo, AICargo.CC_PASSENGERS);
	local stationType = (truckStation) ? AIStation.STATION_TRUCK_STOP : AIStation.STATION_BUS_STOP;
	local radius = AIStation.GetCoverageRadius(stationType);
	
		foreach(target in service.GetTargets()) {
		if (!target.IsValid()) return;
		stations[target.GetId()] <- RoadManager.GetStations(target, cargo, service.GetSubType());
		if(!target.IsTown()) isIndustry = true;

		// If the engine type is articulated, forbid the vehicle from visiting regular stations
/*		if(AIEngine.IsArticulated(engine)) {
			foreach(station, _ in stations[target.GetId()]) {
				local driveThru = AIRoad.IsDriveThroughRoadStationTile(AIStation.GetLocation(station));
				stations[target.GetId()].SetValue(station, (driveThru) ? 1 : 0);
			}
			stations[target.GetId()].RemoveValue(0);
		}
*/
		// If the target has no stations then there is no point in building a 
		// fleet - defer until stations have been built
		if(stations[target.GetId()].Count() == 0) {
//			AILog.Warning("No stations at " + target.GetName());
			return;
		}
	}
	

// Select an engine type
local funds = min(2500000,max(0, FinanceManager.GetAvailableFunds() - (PathZilla.FLOAT/10)*vnum));
	local engine = null;
local price=1000000
local fz=0;
	if(update) {
for (;funds-price<0;funds=FinanceManager.GetAvailableFunds() - (PathZilla.FLOAT/10)*vnum){
if (fz%10==9){
SaveMoney();
}
PathZilla.Sleep(10);
fz++;
::pz.HandleEvents();
		engine = this.SelectEngine(service.GetTargets(), cargo, service.GetTransportType(), service.GetSubType(), true);;
if (engine!=null){
price = AIEngine.GetPrice(engine);
}else{
price=20000;
}	
}
if (engine == null){
engine=service.GetEngine();
if (engine == null){
(::german)?AILog.Warning("  Geplante Verbindung wird verworfen..."):AILog.Warning("  Demoting service...");
SellVehicles(service,service.GetActualFleetSize());
			service.CloseService();
			return false;
}
}
} else {
for (;funds-price<0;funds=FinanceManager.GetAvailableFunds() - (PathZilla.FLOAT/10)*vnum){
if (fz%10==9){
SaveMoney();
}
PathZilla.Sleep(10);
::pz.HandleEvents();
fz++
engine = this.SelectEngine(service.GetTargets(), cargo, service.GetTransportType(), service.GetSubType(), true);
if (engine!=null){
price = AIEngine.GetPrice(engine);
}else{
price=20000;
}
}
if (engine == null){
engine=service.GetEngine();
if (engine == null){
(::german)?AILog.Warning("  Geplante Verbindung wird verworfen..."):AILog.Warning("  Demoting service...");
			this.potentialServices.Remove(service);
			return false;
}
}
		service.SetEngine(engine);
	}	
// Get a few basic details
local capacity = AIEngine.GetCapacity(engine);
	local speed = AIEngine.GetMaxSpeed(engine);
	local distance = 0;
	local prev = 0;
	local fz=0;
	for(local next = 1; next < service.GetTargets().len(); next++) {
		distance += AITile.GetDistanceManhattanToTile(service.GetTargets()[prev].GetLocation(), service.GetTargets()[next].GetLocation());
	}

	// Calculate the required fleet size
	local fleetSize = 0;
	local updateSize=0;
	local jam=0;
	// If we are updating the service, base the decision on waiting cargo
	if(update) {
		// Get the total waiting cargo for all targets
		local waitingCargo = 0;
		local rating=85;
		local vv=0;
		// Prime the station lists with waiting cargo values
		foreach(target in service.GetTargets()) {
			foreach(station, _ in stations[target.GetId()]) {
				local waiting = AIStation.GetCargoWaiting(station, cargo)+max(0,(60-AIStation.GetCargoRating(station,cargo)));
//				local cap = PathZilla.PAX_SERVICE_CAP_BASE * PathZilla.GetSetting("traffic");
				stations[target.GetId()].SetValue(station, waiting);
				if (fz==0)rating=min(AIStation.GetCargoRating(station,cargo),rating);
				jam+=service.VehiclesWaiting(station,AIStation.GetCargoRating(station,cargo)<1);
				if (AIVehicleList_Station(station).IsEmpty()&&service.GetDistance()>1)vv++;
if (rating==69&&AIVehicleList_Station(station).HasItem(service.GetVehicles().Begin())){
TownManager.BuildStatue(AIStation.GetNearestTown(station));
return false;
}
		}
			if (service.IsTwoWayService()){ fz=0;
}else{ fz++;}
			waitingCargo += ListSum(stations[target.GetId()]);
		}

		// Estimate the number of additional vechiles required based on waiting cargo
/*		local year = AIDate.GetYear(AIDate.GetCurrentDate());
		year = min(max(year, 1915), 1950);
		local multiplier = (70 - (year - 1900)) / 2;
		multiplier /= PathZilla.GetSetting("traffic");*/
if(AICargo.GetTownEffect(cargo)==AICargo.TE_MAIL)waitingCargo=waitingCargo/2;
if (waitingCargo==0&&!((service.GetTargets()[0]).IsTown())){
if (!AIIndustryType.IsRawIndustry(AIIndustry.GetIndustryType((service.GetTargets()[0]).GetId())))service.VehiclesWaiting((stations[(service.GetTargets()[0]).GetId()]).Begin(),true);
}
if (waitingCargo<capacity/2&&(rating>65||service.IsTwoWayService())){
if (jam-min(3,((service.GetActualFleetSize()+1)/(AIVehicleList_Station((stations[(service.GetTargets()[0]).GetId()]).Begin()).Count()+1)*service.GetActualFleetSize()))-(service.GetActualFleetSize()/10).tointeger()>0){
		updateSize = service.GetActualFleetSize()-min((service.GetActualFleetSize()/3+1).tointeger(),(jam-min(3,((service.GetActualFleetSize()+1)/(AIVehicleList_Station((stations[(service.GetTargets()[0]).GetId()]).Begin()).Count()+1)*service.GetActualFleetSize()))-(service.GetActualFleetSize()/20).tointeger()));
}else{
updateSize = service.GetActualFleetSize()+(1-min(max(0,jam-(service.GetActualFleetSize()/25).tointeger()),1))*(55-max(54,min(rating,55)))*min(1,service.GetDistance()-1)*(1-min(1,service.IsProfitable(true)))*((service.IsLocked())?0:1)+vv;
}
}else{
		updateSize = (service.GetActualFleetSize()+((sqrt(waitingCargo/(capacity*10))).tointeger()+(1-min(max(0,jam-(service.GetActualFleetSize()/25).tointeger()),1))*(55-max(54,min(rating,55))))*min(1,service.GetDistance()-1)*(1-min(1,service.IsProfitable(true))))*((service.IsLocked())?0:1)+vv;
}
	}
if (service.IsTwoWayService()) jam=(jam/max(3,service.GetActualFleetSize()/10)).tointeger();

	// Find the minimum acceptance level
	local minAcceptance = 1000000;
	local accSum = {};
local prod=0;
local zz=0;
local tt=0;
local big=0;
local stc=0;
	foreach(target in service.GetTargets()) {
		foreach(station, _ in stations[target.GetId()]) {
	
			local acceptance = (AITile.GetCargoAcceptance(AIStation.GetLocation(station), cargo, 1, 1, radius) + 1)*(100-target.GetTransportation(station,cargo))/100;

if (zz==0){
stations[target.GetId()].SetValue(station, 100-AIStation.GetCargoRating(station,cargo));
}else{
			stations[target.GetId()].SetValue(station, acceptance*1000/(AIVehicleList_Station(station).Count()+1));
}
if (zz==0||(service.IsTwoWayService()&&prod*tt<target.GetProduction(station,cargo)*max(40,target.GetTransportation(station,cargo)))){
big=AITile.GetClosestTown(target.GetLocation());
prod=target.GetProduction(station,cargo);
tt=100-max(40,target.GetTransportation(station,cargo));
stc=(stations[target.GetId()]).Count()
}
		}
zz++
	// Get the minimum acceptance of all targets
	accSum[target.GetId()] <- ListSum(stations[target.GetId()]);
if(target.AcceptsCargo(cargo)) {
	minAcceptance = min(minAcceptance, accSum[target.GetId()]);
}
		}
	if (minAcceptance==1000000) minAcceptance=0;
	// Estimate the amount that will be waiting and other details that will 
	// influence our decision on fleet size.
		// Estimate how many vehicles will be needed to cover the route
	if(isIndustry) {
fleetSize = ((((PathZilla.GetSetting("traffic") * prod* (distance+3-(((distance*distance)/2000).tointeger()))*tt) / capacity) / speed)/200+1)*PathZilla.INDUSTRY_FLEET_MULTI/2;
}else{
fleetSize = ((((PathZilla.GetSetting("traffic") * prod* (distance+3-(((distance*distance)/2000).tointeger()))*tt) / capacity) / speed)/200+1)/stc;
}			
//AILog.Warning(PathZilla.GetSetting("traffic")*prod/2+ "*"+tt+"% /" + capacity +"*"+(distance+3-(((distance*distance)/2000).tointeger())) +"/"+speed+"="+fleetSize);
	// Adjust the fleet size for early routes
	// TODO: Make this more generic
//	local year = AIDate.GetYear(AIDate.GetCurrentDate());
//	if(year < 1950) {
//		year = max(year, 1905);
//		fleetSize = (fleetSize * (year - 1900)) / 50;
//	}

	// Ensure the fleet is not too small, there is at least one vehicle per station
	local minFleetSize = 0;
	foreach(target in service.GetTargets()) {
		minFleetSize += stations[target.GetId()].Count();
	}
	fleetSize = max(minFleetSize, fleetSize);

	if(update) {
if (min((service.GetActualFleetSize()/3+1).tointeger(),jam-(min(3,((service.GetActualFleetSize()+1)/(AIVehicleList_Station((stations[(service.GetTargets()[0]).GetId()]).Begin()).Count()+1)*service.GetActualFleetSize()))+(service.GetActualFleetSize()/10).tointeger()))>0&&(fleetSize<service.GetActualFleetSize()||service.IsTwoWayService())) {
SellVehicles(service,min((service.GetActualFleetSize()/3+1).tointeger(),jam-(min(3,((service.GetActualFleetSize()+1)/(AIVehicleList_Station((stations[(service.GetTargets()[0]).GetId()]).Begin()).Count()+1)*service.GetActualFleetSize()))+(service.GetActualFleetSize()/10).tointeger())),true);
return;
}
if (!(service.GetTargets()[0]).IsTown()||jam-1-(service.GetActualFleetSize()/10).tointeger()>0){
fleetSize = min(max(updateSize,service.GetActualFleetSize()),((fleetSize*3)/2).tointeger());
}else{
fleetSize = min(max(updateSize,service.GetActualFleetSize()),max(((fleetSize*3)/2).tointeger(),service.GetActualFleetSize()+1))
}
}

	
	// Ensure we have no mare than the maximum number of vehicles
	fleetSize = min(fleetSize, PathZilla.MAX_VEHICLES_PER_SVC);
	
	// If were updating, account for vehicles already built
	if(update) {
		updateSize = fleetSize - service.GetActualFleetSize();
		local zz=service.IsProfitable();
		if(zz>0) {
(::german)?AILog.Warning("Transport von " + service + " erbringt keinen Profit."):AILog.Warning("Service for " + service + " did not turn a profit last year");
local ee = RoadManager.BuildInfrastructure(service, ::pz.GetSchema(service.GetSchemaId()), this.targetsUpdated);
if (ee==false||zz>=service.GetVehicles().Count()){
SellVehicles(service,service.GetVehicles().Count());
(::german)?AILog.Warning("  Verbindung wird geschlossen..."):AILog.Warning("  Demoting service...");
service.CloseService();
return;	
}else{		
SellVehicles(service,max(zz,1))
			return;
}		
} else {
			fleetSize = max(0, updateSize);
		}
	}else{
if (service.IsTwoWayService()){
fleetSize=fleetSize/((isIndustry)?max(stations[(service.GetTargets()[0]).GetId()].Count(),stations[(service.GetTargets()[1]).GetId()].Count()):max(max(20,AITown.GetHouseCount((service.GetTargets()[0]).GetId()-1000000))/20,max(20,AITown.GetHouseCount((service.GetTargets()[1]).GetId()-1000000))/20)).tointeger();
}else{
fleetSize=fleetSize/(((isIndustry)?stations[(service.GetTargets()[0]).GetId()].Count():max(20,AITown.GetHouseCount(target.GetId()-1000000))/20)).tointeger();
}
fleetSize=max(1,fleetSize);
}
	
	// Do not attempt to build more than we can actually afford
	
	local maxFleetSize = funds / AIEngine.GetPrice(engine);
	fleetSize = min(maxFleetSize, fleetSize);

	// If there is no fleet to build then just return now
	if(fleetSize <= 0) return;
local nodep=false;
local bhff={};
local stz=0;
		foreach(target in service.GetTargets()) {
stz++;
RoadManager.BuildIndustryStation(target, service.GetCargo(), service.GetSubType(),fleetSize);
stations[target.GetId()]<-RoadManager.GetStations(target, cargo, service.GetSubType());
local conmax=0;
local cons=stations[target.GetId()].Begin();
foreach(stdd,_ in RoadManager.GetStations(target, cargo, service.GetSubType())){
local elist=AIVehicleList_Station(stdd);
local empty=true;
foreach(v,_ in elist){
if (AIEngine.GetCargoType(AIVehicle.GetEngineType(v))==AIEngine.GetCargoType(engine)){
empty=false;
break;
}
}
if (empty&&(stz==1||service.IsTwoWayService())){
conmax=100000000;
cons=stdd;
break;
}else if (100-AIStation.GetCargoRating(stdd,cargo)>conmax&&(stz==1||service.IsTwoWayService())){
conmax=100-AIStation.GetCargoRating(stdd,cargo);
cons=stdd;
}else if (AIStation.GetConstructionDate(stdd)>conmax&&stz==2&&!service.IsTwoWayService()){
conmax=AIStation.GetConstructionDate(stdd);
cons=stdd;
}
stations[target.GetId()].SetValue(stdd, 0)
}
stations[target.GetId()].SetValue(cons, accSum[target.GetId()])
bhff[target.GetId()]<-RandomItemByWeight(stations[target.GetId()], accSum[target.GetId()]);
	local depotList = AIDepotList(AITile.TRANSPORT_ROAD);
	depotList.Valuate(AIRoad.HasRoadType, service.GetSubType());
	depotList.KeepValue(1);
	depotList.Valuate(AITile.GetDistanceManhattanToTile, AIStation.GetLocation(bhff[target.GetId()]));
	depotList.Sort(AIAbstractList.SORT_BY_VALUE, false);
	depotList.KeepBottom(1);
	depots[target.GetId()] <- depotList.Begin();
(AITile.GetDistanceManhattanToTile(AIStation.GetLocation(bhff[target.GetId()]),depots[target.GetId()])>10)?nodep=true:nodep=false;
}
local newSt=AIVehicleList_Station(bhff[(service.GetTargets()[0]).GetId()]).IsEmpty()
	local engineName = AIEngine.GetName(engine);
(::german)?AILog.Info(((update) ? "   Aufstockung um " : "  Kaufe ") + fleetSize + " " + engineName + "..."):AILog.Info(((update) ? "  Updating a fleet with " : "  Building a fleet of ") + fleetSize + " " + engineName + "s...");
	
	// Borrow enough to buy the whole fleet of vehicles
	FinanceManager.EnsureFundsAvailable(AIEngine.GetPrice(engine) * (fleetSize + 1));
	
	// Check if the vehicles will need to be refitted
	local needRefit = (AIEngine.GetCargoType(engine) != service.GetCargo());
	
	// Clone a fleet from the prototype vehicle
local zz=0;
local ee=0;
local dep=0;
if (nodep||distance<60) {
dep=0;
}else{
dep=1;
}
	for(local i = 0; i < fleetSize; i++) {
		// Wait some time to spread the vechiles out a bit.
PathZilla.Sleep(PathZilla.NEW_VEHICLE_SPREAD_DELAY);
		// Alternate between targets if they are towns 
		local idx = 0; 
		if(service.IsTwoWayService()) idx = (i + 1) % service.GetTargets().len();
if (idx>0 && service.GetTargets()[idx]==service.GetTargets()[idx-1]){
(::german)?AILog.Warning("  Kann Ziele nicht finden..."):AILog.Warning("  Demoting service...");
			this.potentialServices.Remove(service);
			return false;
}
	local depot = depots[(service.GetTargets()[idx]).GetId()];
		// Build a new vehicle at the nearest depot
	local v = AIVehicle.BuildVehicle(depot, engine);
	// Choose stations and assign orders
		local j = 0;
local jj=0;		
local bz=0;
zz=zz*ee+1;
local ee=1;
foreach(target in service.GetTargets()) {
local bhf=bhff[target.GetId()];
if ((bz==0&&i>0&&!AIIndustryType.IsRawIndustry(AIIndustry.GetIndustryType(target.GetId()))&&!service.IsTwoWayService()&&AIStation.GetCargoWaiting(bhf, cargo)<AIEngine.GetCapacity(engine)*zz)||(!newSt&&!((service.GetTargets()[0]).IsTown())&&bz==0&&i>0&&AIStation.GetCargoWaiting(bhf, cargo)<AIEngine.GetCapacity(engine)*zz)){
for (local ww=0;ww<60&&AIStation.GetCargoWaiting(bhf, cargo)<AIEngine.GetCapacity(engine)*zz;ww++){
if (ww%60==39){
if (AIVehicle.IsValidVehicle(service.GetVehicles().Begin())){
if (i>5&&AIMap.DistanceManhattan(AIVehicle.GetLocation(service.GetVehicles().Begin()),AIStation.GetLocation(bhf))<5&&AIVehicle.GetCargoLoad(service.GetVehicles().Begin(),cargo)<1) return;
}
service.VehiclesWaiting(bhf,true);
SaveMoney();
}
PathZilla.Sleep(10);
::pz.HandleEvents();
ee=0;
}
}
bz++;
			local tile = AIStation.GetLocation(bhf);
			local flags = AIOrder.AIOF_NON_STOP_INTERMEDIATE | AIOrder.AIOF_FULL_LOAD_ANY;
//			if(!target.IsTown() && target.ProducesCargo(cargo)) flags = flags | AIOrder.AIOF_FULL_LOAD;

if (service.IsTwoWayService()){
if (big==AITile.GetClosestTown(tile)) {
AIOrder.AppendOrder(v, tile, flags);
if (AIMap.IsValidTile(depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()])&&AIRoad.IsRoadDepotTile(depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()])&&dep==1){
AIOrder.AppendOrder(v, depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()], AIOrder.AIOF_NON_STOP_INTERMEDIATE);
}else if (dep==1){
AIOrder.AppendOrder(v, depots[service.GetTargets()[(bz+1)%service.GetTargets().len()].GetId()], AIOrder.AIOF_NON_STOP_INTERMEDIATE);
}
}else {
AIOrder.AppendOrder(v, tile, AIOrder.AIOF_NON_STOP_INTERMEDIATE);
if (AIMap.IsValidTile(depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()])&&AIRoad.IsRoadDepotTile(depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()])&&dep==1){
AIOrder.AppendOrder(v, depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()], AIOrder.AIOF_NON_STOP_INTERMEDIATE);
}else if (dep==1){
AIOrder.AppendOrder(v, depots[service.GetTargets()[(bz+1)%service.GetTargets().len()].GetId()], AIOrder.AIOF_NON_STOP_INTERMEDIATE);
}
}
}else {
 AIOrder.AppendOrder(v, tile, flags);
if (AIMap.IsValidTile(depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()])&&AIRoad.IsRoadDepotTile(depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()])&&dep==1){
AIOrder.AppendOrder(v, depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()], AIOrder.AIOF_NON_STOP_INTERMEDIATE);
}else if (dep==1){
AIOrder.AppendOrder(v, depots[service.GetTargets()[(bz+1)%service.GetTargets().len()].GetId()], AIOrder.AIOF_NON_STOP_INTERMEDIATE);
}
 AIOrder.AppendOrder(v, tile, AIOrder.AIOF_NON_STOP_INTERMEDIATE + AIOrder.AIOF_NO_LOAD + AIOrder.AIOF_UNLOAD);
if (AIMap.IsValidTile(depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()])&&AIRoad.IsRoadDepotTile(depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()])&&dep==1){
AIOrder.AppendOrder(v, depots[service.GetTargets()[bz%service.GetTargets().len()].GetId()], AIOrder.AIOF_NON_STOP_INTERMEDIATE );
}else if (dep==1){
AIOrder.AppendOrder(v, depots[service.GetTargets()[(bz+1)%service.GetTargets().len()].GetId()], AIOrder.AIOF_NON_STOP_INTERMEDIATE);
}
 AIOrder.InsertConditionalOrder(v, jj,jj+1+dep);                        
 AIOrder.SetOrderCondition(v, jj, AIOrder.OC_LOAD_PERCENTAGE);
 AIOrder.SetOrderCompareFunction(v, jj, AIOrder.CF_MORE_THAN);
 AIOrder.SetOrderCompareValue(v, jj, 0);
if (jj>0) { AIOrder.InsertConditionalOrder(v, jj-1-dep,jj);                        
 AIOrder.SetOrderCondition(v, jj-1-dep, AIOrder.OC_UNCONDITIONALLY);
jj++;
}
		jj=jj+3+dep*2;	
		}
}
if (jj>0 && !service.IsTwoWayService()) { AIOrder.InsertConditionalOrder(v, jj-1-dep,0);                        
 AIOrder.SetOrderCondition(v, jj-1-dep, AIOrder.OC_UNCONDITIONALLY);
}		
		// Refit the vehicle if necessary
		if(needRefit) {
			AIVehicle.RefitVehicle(v, service.GetCargo());
		}
		// Send the vehicle to the destination nearest the depot we built it at
		AIVehicle.SkipToVehicleOrder(v, idx);
		
		// Add the vehicle to the service
		service.AddVehicle(v);

		// Start the vehicle
		AIVehicle.StartStopVehicle(v);
	}
}

function ServiceManager::SaveMoney() {
foreach(service in this.serviceList) {
local zfz=0;
(AIDate.GetMonth(AIDate.GetCurrentDate())<8)?zfz=service.IsProfitable():zfz=service.IsProfitable(true);
if (zfz>0){
SellVehicles(service,(zfz/9).tointeger()+1);
PathZilla.Sleep(100);
}else{
local smjam=0;
local sfz=0;
local smcargo=service.GetCargo();
local smstations={}
local smwaitingCargo=0;
foreach(smtarget in service.GetTargets()) {
	smstations[smtarget.GetId()] <- RoadManager.GetStations(smtarget, smcargo, service.GetSubType());
if ((smstations[smtarget.GetId()]).Begin()!=null){
	foreach(smstation, _ in smstations[smtarget.GetId()]) {
		local smwaiting = AIStation.GetCargoWaiting(smstation, smcargo)+max(0,(50-AIStation.GetCargoRating(smstation,smcargo)));
		smstations[smtarget.GetId()].SetValue(smstation, smwaiting);
		if (sfz==0){
		smjam+=service.VehiclesWaiting(smstation);
		sfz++
		}
	}
	if (service.IsTwoWayService()) sfz=0;
	smwaitingCargo = ListSum(smstations[smtarget.GetId()]);
}
if (service.IsTwoWayService()) smjam=smjam/2;
if (service.GetActualFleetSize()<1||(smstations[(service.GetTargets()[0]).GetId()]).Count()==0||AIVehicleList_Station((smstations[(service.GetTargets()[0]).GetId()]).Begin()).IsEmpty()) continue;
if (smwaitingCargo<AIEngine.GetCapacity(AIVehicle.GetEngineType(service.GetVehicles().Begin()))/2&&min((service.GetActualFleetSize()/2+1).tointeger(),smjam-(((min(3,((service.GetActualFleetSize()+1)/(AIVehicleList_Station((smstations[(service.GetTargets()[0]).GetId()]).Begin()).Count()+1)*service.GetActualFleetSize()))+(service.GetActualFleetSize()/10).tointeger())*2)/3).tointeger())>0){
service.VehiclesWaiting((smstations[smtarget.GetId()]).Begin(),true);
SellVehicles(service,min((service.GetActualFleetSize()/2+1).tointeger(),smjam-(((min(3,((service.GetActualFleetSize()+1)/(AIVehicleList_Station((smstations[(service.GetTargets()[0]).GetId()]).Begin()).Count()+1)*service.GetActualFleetSize()))+(service.GetActualFleetSize()/10).tointeger())*2)/3).tointeger()),true);
}
smjam=0;
}
}
}
local sz=0;
for (;AICompany.GetBankBalance(AICompany.COMPANY_SELF)*3+1000<PathZilla.FLOAT*2;){

for (;AICompany.GetBankBalance(AICompany.COMPANY_SELF)<0;){
sz++;
local minprofit=10000000;
local sell=null;
if (::vehiclesToSell.Count()==0||sz%100==99){
foreach(service in this.serviceList) {
foreach (v,_ in service.GetVehicles()){
if (AIVehicle.IsValidVehicle(v)){
if (AIVehicle.GetAge(v)>400&&AIVehicle.GetProfitLastYear(v)<minprofit){
minprofit=AIVehicle.GetProfitLastYear(v);
sell=service;
}
}
}
}
}
if (minprofit<0){
SellVehicles(sell,1);
}
PathZilla.Sleep(50);
::pz.HandleEvents();
}
PathZilla.Sleep(1);
::pz.HandleEvents();
}
return;
}

/*
 * Sell the specified number of vehicles from the specified service, selected 
 * at random. This will permanently reduce the service's fleet size.
 */
function ServiceManager::SellVehicles(service, number,jam=false) {
if (service.GetActualFleetSize()>0){
service.Lock((service.GetActualFleetSize()/(service.GetActualFleetSize()+1-number))*300)
local gid=AIVehicle.GetGroupID(service.GetVehicles().Begin());
foreach (v,_ in ::vehiclesToSell){
if (!AIVehicle.IsValidVehicle(v)) continue;
if (AIVehicle.GetGroupID(v)==gid)number--;
}
}else return;
if (number <1) return;
if (jam)((::german)?AILog.Warning("Stau beim Transport von "+service+ "."):AILog.Warning("Service for "+service+ ": vehicles are jamming"));
(::german)?AILog.Info("  Verkaufe " + number + " Fahrzeug" +((number>1)?"e":"")+"..."):AILog.Info("  Selling " + number + " vehicles...");
	local vlist = service.GetVehicles();
	vlist.Valuate(function(veh,svc){return (svc.VehiclesWaiting(AIStationList_Vehicle(veh).Begin())+svc.VehiclesWaiting(AIStationList_Vehicle(veh).Next()))*1000+AIVehicle.GetAge(veh);},service);
	foreach(vehicle, _ in vlist) {
		local empty = (AIVehicle.GetCargoLoad(vehicle, service.GetCargo()) == 0);
		if (!empty) vlist.SetValue(vehicle, vlist.GetValue(vehicle)*10*AIVehicle.GetCargoLoad(vehicle, service.GetCargo()));
	}
	vlist.Sort(AIAbstractList.SORT_BY_VALUE, false);
	
	// Clone a fleet from the prototype vehicle
	local i = 1;
	for(local vehicle = vlist.Begin(); hasnext(vlist) && i++ <= number; vehicle = vlist.Next()) {

		// Turn the vehcile around (to help clear jams) and send it to a depot
		AIVehicle.SendVehicleToDepot(vehicle);
		AIVehicle.ReverseVehicle(vehicle);
		
		// Remember to sell the vehcile when it stops in a depot
		::vehiclesToSell.AddItem(vehicle, 0);
service.RemoveVehicle(vehicle);	
}
}
/*
 * Update orders for all the vehicles in a service to ensure that vechicles are
 * distributed correctly between the available stations. 
 */
function ServiceManager::UpdateOrders(service) {
	// Get the stations
if (service.GetDistance()==1) return;
	local stations = {};
	foreach(target in service.GetTargets()) {
		stations[target.GetId()] <- RoadManager.GetStations(target, cargo, service.GetSubType());

		// If the engine type is articulated, forbid the vehicle from visiting regular stations
		if(AIEngine.IsArticulated(engine)) {
			foreach(station, _ in stations[target.GetId()]) {
				local driveThru = AIRoad.IsDriveThroughRoadStationTile(AIStation.GetLocation(station));
				stations[target.GetId()].SetValue(station, (driveThru) ? 1 : 0);
			}
			stations[target.GetId()].RemoveValue(0);
		}

		// If the target has no stations then there is no point in building a 
		// fleet - defer until stations have been built
		if(stations[target.GetId()].Count() == 0) {
(::german)?AILog.Warning("Keine Haltestelle in " + target.GetName()):AILog.Warning("No stations at " + target.GetName());
			return;
		}
	}

	// Get the coverage radius of the stations	
	local truckStation = !AICargo.HasCargoClass(service.GetCargo(), AICargo.CC_PASSENGERS);
	local stationType = (truckStation) ? AIStation.STATION_TRUCK_STOP : AIStation.STATION_BUS_STOP;
	local radius = AIStation.GetCoverageRadius(stationType);

	// Find the acceptance list sums
	local accSum = {};
	
	foreach(target in service.GetTargets()) {
		foreach(station, _ in stations[target.GetId()]) {
			local acceptance = AITile.GetCargoAcceptance(AIStation.GetLocation(station), cargo, 1, 1, radius) + 1;
			stations[target.GetId()].SetValue(station, acceptance);
		}

		// Get the minimum acceptance of all targets
		accSum[target.GetId()] <- ListSum(stations[target.GetId()]);
	}

	// Shuffle the vehicle orders between the stations
	foreach(v, _ in service.GetVehicles()) {
		local currentOrder = AIOrder.ResolveOrderPosition(v, AIOrder.ORDER_CURRENT);

		// Clear the order list
		local ocount = AIOrder.GetOrderCount(v);
		for(local i = 0; i < ocount; i++) {
			AIOrder.RemoveOrder(v, i);
		}

		// Set the new orders
		foreach(target in service.GetTargets()) {
			local tile = AIStation.GetLocation(RandomItemByWeight(stations[target.GetId()], accSum[target.GetId()]));
			local flags = AIOrder.AIOF_NON_STOP_INTERMEDIATE;
			if(!target.IsTown() && target.ProducesCargo(cargo)) flags = flags & AIOrder.AIOF_FULL_LOAD;
			AIOrder.AppendOrder(v, tile, flags);
		}

		// Ensure the vehicle is still heading to the same town it was before
		AIVehicle.SkipToVehicleOrder(v, currentOrder);
	}
}

function ServiceManager::profit(a=null,b=null){
if (a==null||b==null) return 0;
if (a.rawIncome==b.rawIncome)return 0;
if (a.rawIncome<b.rawIncome)return 1;
return-1;
}

/*
 * Saves data to a table.
 */
function ServiceManager::Serialize() {
	local data = {};
	
	data[SRLZ_TARGETS_UPDATED] <- this.targetsUpdated.Serialize(); 
	data[SRLZ_TARGETS_CONSIDERED] <- this.targetsConsidered.Serialize(); 
	data[SRLZ_POTENTIAL_SERVICES] <- this.potentialServices.Serialize();
	data[SRLZ_SERVICE_LIST] <- this.serviceList.Serialize();
	
	return data;
}

/*
 * Loads data from a table.
 */
function ServiceManager::Unserialize(data) {
	this.targetsUpdated = Map();
	this.targetsUpdated.Unserialize(data[SRLZ_TARGETS_UPDATED]); 

	this.targetsConsidered = Map();
	this.targetsConsidered.Unserialize(data[SRLZ_TARGETS_CONSIDERED]); 
	
	this.potentialServices = SortedSet();
	this.potentialServices.Unserialize(data[SRLZ_POTENTIAL_SERVICES]);

	this.serviceList = SortedSet();
	this.serviceList.Unserialize(data[SRLZ_SERVICE_LIST]);
}

/*
 * This call should be made after data has been loaded and the game has 
 * started, to load vehicles into the service list.
 */
function ServiceManager::PostLoad() {
	foreach(service in this.serviceList) {
		local vehicles = AIVehicleList();
		vehicles.Valuate(AIVehicle.GetGroupID);
		vehicles.KeepValue(service.group);
		if(vehicles.Count() > 0) service.vehicles.AddList(vehicles);
	}
}
